from typing import Optional, Tuple, Callable, Any

from anygrad.tensor.base import tensor_c as C
from anygrad.tensor.base import ThHelper as Th

import anygrad
import anygrad.autograd as Ag


class BaseTensor:

    _dtype_map = {
        "float32": Th.float32,
        "float64": Th.float64,
        "int32": Th.int32,
        "int64": Th.int64,
        "bool": Th.bool,
    }

    promotion_table = {
        (Th.bool, Th.bool) : Th.bool,
        
        (Th.int32, Th.int32): Th.int32,
        (Th.int32, Th.int64):Th.int64,
        (Th.int32, Th.float32):Th.float32,
        (Th.int32, Th.float64):Th.float64,
        
        (Th.int64, Th.int64): Th.int64,
        (Th.int64, Th.float32):Th.float32,
        (Th.int64, Th.float64):Th.float64,
        
        (Th.float32, Th.float32):Th.float32,
        (Th.float32, Th.float64):Th.float64,
        
        (Th.float64, Th.float64):Th.float64
    }
    
    @classmethod
    def promote_type(cls, type1, type2):
        return cls.promotion_table.get((type1, type2)) or cls.promotion_table.get((type2, type1))

    def __init__(self, requires_grad=False):

        self.requires_grad = requires_grad and Ag.GradMode.is_enabled()
        self.grad = None
        self.name_backward = ""
        self._backward = lambda: None
        self._prev = set()
        self.is_leaf = True

    @property
    def shape(self) -> Tuple[int, ...]:
        return tuple(self.base.shape)

    @property
    def ndim(self) -> int:
        return self.base.ndim

    @property
    def size(self) -> int:
        return self.base.size

    @property
    def dtype(self) -> str:
        return BaseTensor._dtype_map[self.base.dtype]
    
    @staticmethod
    def _create_tensor(ResultClass, OtherClass, allow_other_class, data, shape, dtype, req_grad):
        reshape = Th.reshape(data, shape)
        
        if allow_other_class:
            ans = OtherClass(reshape, requires_grad=req_grad, dtype=dtype)
        elif dtype in {Th.float32, Th.float64}:
            ans = OtherClass(reshape, requires_grad=req_grad, dtype=dtype)
        elif dtype in {Th.int32, Th.int64, Th.bool}:
            ans = ResultClass(reshape, dtype=dtype)
        return ans

    @staticmethod
    def _apply_operation(tensor1, 
                         tensor2, 
                         ResultClass, 
                         OtherClass, 
                         has_scaler:bool, 
                         operation:Callable, 
                         operation_name:str,
                         allow_other_class:bool, 
                         broadcast_check:Callable = C.isbroadcast):
        
        if isinstance(tensor2, (int, float)) and has_scaler:
            data = [operation(i, tensor2) for i in tensor1.base.data]
            ans = BaseTensor._create_tensor(ResultClass, OtherClass, allow_other_class, data, tensor1.base.shape, tensor1.dtype, tensor1.requires_grad)
            
            if ans.requires_grad:
                ans._prev = {tensor1}
                ans._backward = getattr(Ag.GradientCal, f"{operation_name.capitalize()}_grad")(tensor1, tensor2, ans)
                ans.name_backward = f"<{operation_name}Backward1>"
                ans.is_leaf = False
            return ans
        
        allow = broadcast_check(tensor1.base.shape, tensor2.base.shape, tensor1.base.ndim, tensor2.base.ndim)
        if not allow:
            raise RuntimeError(f"The size of the tensors are must broadcast current shapes are :{tensor1.base.shape} and {tensor2.base.shape}")
        
        func = getattr(C, f"{operation_name.capitalize()}", None)
        if func is None:
            raise NotImplementedError(f"""
                                      {operation_name} is not implemented in C++
                                      """)
        
        data, shape = func(tensor1.base, tensor2.base)
        dtype = BaseTensor.promote_type(tensor1.dtype, tensor2.dtype)
        req_grad = tensor1.requires_grad or tensor2.requires_grad
        ans = BaseTensor._create_tensor(ResultClass, OtherClass, allow_other_class, data, shape, dtype, req_grad)
        if ans.requires_grad:
            ans._prev = {tensor1, tensor2}
            ans._backward = getattr(Ag.GradientCal, f"{operation_name.capitalize()}_grad")(tensor1, tensor2, ans)
            ans.name_backward = f"<{operation_name}Bacward0>"
            ans.is_leaf = False
            
        return ans

    @staticmethod
    def _reduce_ops(tensor1, ResultClass, OtherClass, axis: Optional[int], keepdims: bool, operation_name: str, allow_other_class: bool = False):
        allow = C.is_sum_allow(axis, tensor1.base.ndim)
        if not allow:
            raise RuntimeError(f"Invalid reduction operation: axis {axis} is not compatible with tensor of dimension {tensor1.base.ndim}")

        func = getattr(C, f"{operation_name.capitalize()}", None)
        if func is None:
            raise NotImplementedError(f"{operation_name} is not implemented in C++")

        data, shape = func(tensor1.base, axis, keepdims)
        ans = BaseTensor._create_tensor(ResultClass, OtherClass, allow_other_class, 
                                      data, shape, tensor1.dtype, tensor1.requires_grad)

        if ans.requires_grad:
            ans._prev = {tensor1}
            ans._backward = getattr(Ag.GradientCal, f"{operation_name.capitalize()}_grad")(
                tensor1, ans
            )
            ans.name_backward = f"<{operation_name}Backward0>"
            ans.is_leaf = False

        return ans

    @staticmethod
    def _trans_ops(tensor1, ResultClass, OtherClass, dim0: int, dim1: int, allow_other_class: bool = False):
        if dim0 < 0 and dim1 < 0:
            dim0 = tensor1.ndim + dim0
            dim1 = tensor1.ndim + dim1

        if tensor1.ndim < 2:
            raise RuntimeError(f"Transpose requires at least 2D tensor, got {tensor1.ndim}D")
        
        func = getattr(C, f"Trans", None)
        if func is None:
            raise NotImplementedError("Transpose is not implemented in C++")
            
        data, shape = func(tensor1.base, dim0, dim1)
        ans = BaseTensor._create_tensor(ResultClass, OtherClass, allow_other_class,
                                      data, shape, tensor1.dtype, tensor1.requires_grad)

        if ans.requires_grad:
            ans._prev = {tensor1}
            ans.name_backward = "<TransBackward0>"
            ans._backward = getattr(Ag.GradientCal, "Trans_grad")(tensor1, ans)
            ans.is_leaf = False

        return ans
    
    @staticmethod
    def _apply_view(tensor1, ResultClass, OtherClass, shape, allow_other_class: bool = False):
        allow = C.is_view_allow(tensor1.base.shape, tensor1.base.size)
        if not allow:
            raise RuntimeError(f"Invalid view operation: shape {shape} is not compatible with tensor of size {tensor1.base.size}")
        
        func = getattr(C, f"View", None)
        if func is None:
            raise NotImplementedError("View is not implemented in C++")
        
        data, shape = func(tensor1.base, shape)
        ans = BaseTensor._create_tensor(ResultClass, OtherClass, allow_other_class,
                                      data, shape, tensor1.dtype, tensor1.requires_grad)
        
        if ans.requires_grad:
            ans._prev = {tensor1}
            ans.name_backward = "<ViewBackward0>"
            ans._backward = getattr(Ag.GradientCal, "View_grad")(tensor1, ans)
            ans.is_leaf = False
            
        return ans

    @staticmethod
    def _apply_reshape(tensor, shape, TensorClass):
        ans = BaseTensor._create_tensor(TensorClass, TensorClass, False,
                                      tensor.data, shape, tensor.dtype, tensor.requires_grad)
        
        if ans.requires_grad:
            ans._prev = {tensor}
            ans.name_backward = "<ReshapeBackward0>"
            ans._backward = getattr(Ag.GradientCal, "Reshape_grad")(tensor, ans)
            ans.is_leaf = False
        
        return ans
    
    @staticmethod
    def _apply_compare(tensor1, tensor2, ResultClass, operation, operation_name:str, has_scaler:bool, broadcast_check = C.isbroadcast):
        
        if isinstance(tensor2, (int, float)) and has_scaler:
            data, shape = [operation(tensor1.data[i], tensor2) for i in range(tensor1.size)]
            reshape = Th.reshape(data, shape)
            ans = ResultClass(reshape, dtype=Th.bool)
            return ans
        
        allow = broadcast_check(tensor1.base.shape, tensor2.base.shape, tensor1.base.ndim, tensor2.base.ndim)
        if not allow:
            raise RuntimeError(f"The size of the tensors are must broadcast current shapes are :{tensor1.base.shape} and {tensor2.base.shape}")
        
        func = getattr(C, f"{operation_name.capitalize()}", None)
        if func is None:
            raise NotImplementedError(f"""
                                      {operation_name} is not implemented in C++
                                      """)
        
        data, shape = func(tensor1.base, tensor2.base)
        reshape = Th.reshape(data, shape)
        ans = ResultClass(reshape, dtype=Th.bool)
        return ans
        
        
        
        
    def __iter__(self):
        return iter(self.data)

    def __neg__(self):
        return -1 * self

    def backward(self, custom_grad=None) -> None:
        if not self.requires_grad:
            raise ValueError("Backward pass only works if requires_grad is True")

        if self.shape == (1,):
            if custom_grad is not None:
                raise ValueError(
                    "Do not provide a custom gradient for scalar outputs; use a scalar value for grad computation"
                )
            self.grad = anygrad.ones_like(self, requires_grad=False, dtype=self.dtype)
        else:
            if custom_grad is None:
                raise ValueError(
                    "A custom gradient must be provided for non-scalar outputs"
                )
            if custom_grad.shape != self.shape:
                raise ValueError(
                    f"Custom grad shape {custom_grad.shape} doesn't match output shape {self.shape}"
                )
            self.grad = custom_grad

        topo = Ag.BuildGraph.construct_graph(self)

        for v in reversed(topo):
            if v is not self and v._prev:
                v.grad = None

        for v in topo:
            v._backward()



class BuildGraph:
    @staticmethod
    def construct_graph(tensor):

        topo = []
        visited = set()
        queue = [tensor]
        while queue:
            ele = queue.pop(0)
            if ele not in visited:
                visited.add(ele)
                topo.append(ele)

                for child in ele._prev:
                    queue.append(child)
        return topo

    __module__ = "anygrad.autograd"



from anygrad.autograd.autograd import GradMode, no_grad
from anygrad.autograd.helper import BuildGraph
from anygrad.autograd.helper import GradientCal

__all__ = ["GradMode", "GradientCal", "BuildGraph", "no_grad"]



from anygrad.utils.generator import Generator, rand, randint
from anygrad.tensor.tensor import Tensor
from anygrad.tensor.base.basetensor import BaseTensor
from anygrad.tensor.base import ThHelper as Th
from anygrad.tensor.tensor import Tensor
from anygrad.utils import utils_c as C

from typing import Tuple, Optional
import anygrad

def _get_dtype_str(dtype):
    for key, val in BaseTensor._dtype_map.items():
        if val == dtype:
            return key.capitalize()

def _use_ops_zeros_ones(shape: tuple, requires_grad: bool, dtype: any, operation_name: str):
    try:
        operation_func = getattr(
            C,
            f"{operation_name.capitalize()}{_get_dtype_str(dtype)}"
        )
    except Exception as e:
        raise NotImplementedError(f"{operation_name} is not Imlemented for {dtype}.") from e

    data, shape = operation_func(shape)
    ans = Th.reshape(data, shape)
    ans = Tensor(ans, dtype=dtype, requires_grad=requires_grad)
    return ans


def _use_ops_log(tensor1, requires_grad: bool, operation_name: str):
    try:
        op_func = getattr(
            C, f"{operation_name.capitalize()}{tensor1.base.dtype.capitalize()}"
        )
    except Exception as e:
        raise NotImplementedError(f"{operation_name} is not Imlemented for {tensor1.base.dtype}.") from e

    data, shape = op_func(tensor1.base)
    ans = Th.reshape(data, shape)
    if tensor1.dtype == Th.int32:
        ans = Tensor(ans, dtype=Th.float32, requires_grad=requires_grad)
    elif tensor1.dtype == Th.int64:
        ans = Tensor(ans, dtype=Th.float64, requires_grad=requires_grad)
    else:
        ans = Tensor(ans, dtype=tensor1.base.dtype, requires_grad=requires_grad)
    return ans

def zeros(
    *shape: Tuple[int],
    requires_grad: Optional[bool] = False,
    dtype: Optional[anygrad.float32 | anygrad.float64] = anygrad.float32,
) -> Tensor:
    return _use_ops_zeros_ones(
        shape=shape, requires_grad=requires_grad, dtype=dtype, operation_name="Zeros"
    )


def ones(
    *shape: Tuple[int],
    requires_grad: Optional[bool] = False,
    dtype: Optional[anygrad.float32 | anygrad.float64] = anygrad.float32,
) -> Tensor:
    return _use_ops_zeros_ones(
        shape=shape, requires_grad=requires_grad, dtype=dtype, operation_name="Ones"
    )


def zeros_like(
    tensor: Tensor,
    requires_grad: Optional[bool] = False,
    dtype: Optional[anygrad.float32 | anygrad.float64] = anygrad.float32,
) -> Tensor:
    requires_grad = (
        requires_grad if requires_grad is not False else tensor.requires_grad
    )
    return zeros(*tensor.shape, requires_grad=requires_grad, dtype=dtype)


def ones_like(
    tensor: Tensor,
    requires_grad: Optional[bool] = False,
    dtype: Optional[anygrad.float32 | anygrad.float64] = anygrad.float32,
) -> Tensor:
    requires_grad = (
        requires_grad if requires_grad is not False else tensor.requires_grad
    )
    return ones(*tensor.shape, requires_grad=requires_grad, dtype=dtype)


def log(tensor: Tensor, requires_grad: Optional[bool] = False) -> Tensor:
    return _use_ops_log(tensor, requires_grad=requires_grad, operation_name="Log")


def log10(tensor: Tensor, requires_grad: Optional[bool] = False) -> Tensor:
    return _use_ops_log(tensor, requires_grad=requires_grad, operation_name="Log10")


def log2(tensor: Tensor, requires_grad: Optional[bool] = False) -> Tensor:
    return _use_ops_log(tensor, requires_grad=requires_grad, operation_name="Log2")


def exp(tensor: Tensor, requires_grad: Optional[bool] = False) -> Tensor:
    return _use_ops_log(tensor, requires_grad=requires_grad, operation_name="Exp")


def exp2(tensor: Tensor, requires_grad: Optional[bool] = False) -> Tensor:
    return _use_ops_log(tensor, requires_grad=requires_grad, operation_name="Exp2")


zeros.__module__ = "anygrad"
ones.__module__ = "anygrad"
ones_like.__module__ = "anygrad"
zeros_like.__module__ = "anygrad"
log.__module__ = "anygrad"
log10.__module__ = "anygrad"
log2.__module__ = "anygrad"
exp.__module__ = "anygrad"
exp2.__module__ = "anygrad"

__all__ = ["Generator", "rand", "randint"]



from typing import NewType
from anygrad.tensor.base import tensor_c as C
from collections.abc import Sequence
from collections import Counter, deque
from functools import reduce
from operator import mul


float32 = NewType("float32", C.float32)
float32.__module__ = "anygrad"
float64 = NewType("float64", C.float64)
float64.__module__ = "anygrad"
int32 = NewType("int32", C.int32)
int32.__module__ = "anygrad"
int64 = NewType("int64", C.int64)
int64.__module__ = "anygrad"
bool = NewType("bool", C.bool)
bool.__module__ = "anygrad"

def convert_tensor(data, conv_type):
    if not isinstance(data, list):
        return conv_type(data)
    return [convert_tensor(ele, conv_type) for ele in data]

def valid_data_type(data):
    flat = flat_list(data) if isinstance(data, list) else [data]
    counter = Counter(type(x) for x in flat)
    return max(counter, key=counter.get)

def flat_list(data):
    if not isinstance(data, list):
        return data
    result = []
    queue = deque([data])
    while queue:
        current = queue.popleft()
        if isinstance(current, list):
            queue.extend(current)
        else:
            result.append(current)
    return result

def cal_shape(data):
    shape = []
    while isinstance(data, Sequence) and not isinstance(data, (str, bytes)):
        if not data:
            shape.append(0)
            break
        if isinstance(data[0], Sequence):
            expected_len = len(data[0])
            if any(len(item) != expected_len for item in data):
                raise ValueError("Not all lists have the same length in your data")
        shape.append(len(data))
        data = data[0]
    return tuple(shape)

def reshape(data, shape):
    
    if reduce(mul, shape) != len(data):
        raise ValueError(f"Given {shape} shape is not compatable with data")
    
    if len(shape) == 1:
        return data

    n = reduce(mul, shape[1:])
    return [reshape(data[i*n: (i + 1)*n], shape[1:]) for i in range(len(data) // n)]


def round_list(data, round_factor=4):
    if isinstance(data, list):
        return [round_list(item, round_factor) for item in data]
    return round(data, round_factor)





"""
    AnyGrad (anygrad)
    ----------
    A Python module that use C++ for Tensor operations.

"""
from anygrad.tensor.tensor import Tensor
from anygrad.tensor.base.floattensor import FloatTensor
from anygrad.tensor.base.inttensor import IntTensor
from anygrad.tensor.base.booltensor import BoolTensor
from anygrad.tensor.base.ThHelper import float32, float64, int32, int64, bool
from anygrad.autograd import no_grad
from anygrad.utils import (Generator, rand, randint, ones, ones_like, zeros, zeros_like, 
                    log, exp, exp2, log10, log2)
from anygrad.version import __version__


def matmul(tensor1, tensor2):
    return tensor1 @ tensor2


def cast(tensor: Tensor, target_dtype):
    return Tensor(tensor.data, requires_grad=tensor.requires_grad, dtype=target_dtype)


__all__ = [
    "Tensor", "FloatTensor", "IntTensor", "BoolTensor", "float32", "float64", "int32", "int64", "bool", "no_grad", 
    "Generator", "rand", "randint", "ones", "ones_like", "zeros", "zeros_like", "log", "log2", "log10", "log2", "exp", "exp2", "matmul", "cast", 
    "__version__",
]



from typing import Optional, List, Tuple, Union

from anygrad.tensor.base import ThHelper as Th
from anygrad.tensor.base.ThHelper import float32
from anygrad.tensor.base.booltensor import BoolTensor
from anygrad.tensor.base.inttensor import IntTensor
from anygrad.tensor.base.floattensor import FloatTensor

from anygrad.version import __version__
from numpy import array


class Tensor:
    """
    Class to repesent a Tensor.

    Attributes
    ----------
    data : List | Tuple
        Any Iterable

    requires_grad: Optional[bool] = False
        if `True` the gradient caculation is happend

    dtype: Optional[anygrad.float32 | anygrad.float64 | anygrad.int32 | anygrad.int64 | anygrad.bool] = anygrad.float32

    Methods
    ----------
    data:
        return the item of the tensor in list form.
    shape:
        return the shape of the tensor in tuple form.
    ndim:
        return the dim of the tensor in int form.
    requires_grad:
        return the bool value if the requires_grad.
    grad:
        a tensor value that allow you to see the gredient of the tensor.

    add(other):
        other: Tensor | int | float
        add the Tensor or number.

    sub(other):
        other: Tensor | int | float
        sub the Tensor or number.

    mul(other):
        other: Tensor | int | float
        mul the Tensor or number.

    div(other):
        other: Tensor | int | float
        div the Tensor or number.

    pow(other):
        other: Tensor | int | float
        pow the Tensor or number.

    matmul(other):
        other: Tensor
        matrix multiplication of the Two valid shape tensor.

    sum(self, axis: Optional[int] = -1, keepdims: Optional[bool] = False) -> Tensor:
        sum the tensor with axis and keepdims.

    backward(self, custom_grad:Optional[Tensor] = None) -> None:
        Do the backward pass if the requires-grad is true for given tensor.

    """

    def __init__(
        self,
        data: List[int | float] | Tuple[int | float],
        requires_grad: Optional[bool] = False,
        dtype:any = float32,
    ) -> None:

        # requires_grad type checking
        if not isinstance(requires_grad, bool):
            raise TypeError(
                f"requires_grad is must be in bool not in {type(requires_grad)}"
            )

        if isinstance(data, (int, float)):
            data = [data]

        if isinstance(dtype, str):
            dtype = getattr(Th, dtype)
        valid_types = {Th.float32, Th.float64, Th.int32, Th.int64, Th.bool}
        if dtype not in valid_types:
            raise TypeError(
                "Tensor must have a valid dtype: 'float32', 'float64', 'int32', 'int64' or 'bool'."
            )

        if dtype not in {Th.float32, Th.float64} and requires_grad:
            raise RuntimeError(
                f"The requires_grad is only support for 'float32' and 'float64' for version {__version__}"
            )

        type_list = Th.valid_data_type(data)
        if dtype == Th.bool:
            self._tensor = BoolTensor(data, dtype=Th.bool)
        elif dtype in {Th.float32, Th.float64}:
            self._tensor = FloatTensor(data, requires_grad=requires_grad, dtype=dtype)
        elif dtype in {Th.int32, Th.int64}:
            self._tensor = IntTensor(data, dtype=dtype)
        else:
            raise TypeError(
                f"Given Datatype {type_list} is not support in current version {__version__}"
            )

    @property
    def shape(self) -> Tuple[int, ...]:
        return self._tensor.shape

    @property
    def ndim(self) -> int:
        return self._tensor.ndim

    @property
    def size(self) -> int:
        return self._tensor.size

    @property
    def dtype(self) -> str:
        return self._tensor.dtype

    @property
    def data(self):
        return self._tensor.data

    @property
    def base(self):
        return self._tensor.base

    @property
    def requires_grad(self):
        return self._tensor.requires_grad

    @requires_grad.setter
    def requires_grad(self, value):
        self._tensor.requires_grad = value

    @property
    def grad(self):
        return self._tensor.grad

    @grad.setter
    def grad(self, value):
        self._tensor.grad = value

    @property
    def _backward(self):
        return self._tensor._backward

    @_backward.setter
    def _backward(self, value):
        self._tensor._backward = value

    @property
    def _prev(self):
        return self._tensor._prev

    @_prev.setter
    def _prev(self, value):
        self._tensor._prev = value

    @property
    def is_leaf(self):
        return self._tensor.is_leaf

    @is_leaf.setter
    def is_leaf(self, value):
        self._tensor.is_leaf = value

    def __repr__(self) -> str:
        return repr(self._tensor)

    def __iter__(self):
        return iter(self.data)

    def __getitem__(self, index: Union[int, slice]) -> "Tensor":
        return self._tensor.__getitem__(index)

    def __add__(self, other: Union["Tensor", int, float]) -> "Tensor":
        return self._tensor.__add__(other)

    def __sub__(self, other: Union["Tensor", int, float]) -> "Tensor":
        return self._tensor.__sub__(other)

    def __mul__(self, other: Union["Tensor", int, float]) -> "Tensor":
        return self._tensor.__mul__(other)

    def __truediv__(self, other: Union["Tensor", int, float]) -> "Tensor":
        return self._tensor.__truediv__(other)

    def __matmul__(self, other: Union["Tensor", int, float]) -> "Tensor":
        return self._tensor.__matmul__(other)

    def __neg__(self) -> "Tensor":
        return self._tensor.__neg__()

    def __pow__(self, other: Union["Tensor", int, float]) -> "Tensor":
        return self._tensor.__pow__(other)

    def __radd__(self, other: Union["Tensor", int, float]) -> "Tensor":
        return self._tensor.__radd__(other)

    def __rsub__(self, other: Union["Tensor", int, float]) -> "Tensor":
        return self._tensor.__rsub__(other)

    def __rmul__(self, other: Union["Tensor", int, float]) -> "Tensor":
        return self._tensor.__rmul__(other)

    def __rtruediv__(self, other: Union["Tensor", int, float]) -> "Tensor":
        return self._tensor.__rtruediv__(other)
    
    def __eq__(self, other:Union["Tensor", int, float]) -> "Tensor":
        return self._tensor.__eq__(other)
    
    def __gt__(self, other:Union["Tensor", int, float]) -> "Tensor":
        return self._tensor.__gt__(other)
    
    def __lt__(self, other:Union["Tensor", int, float]) -> "Tensor":
        return self._tensor.__lt__(other)
    
    def __ge__(self, other:Union["Tensor", int, float]) -> "Tensor":
        return self._tensor.__ge__(other)
    
    def __le__(self, other:Union["Tensor", int, float]) -> "Tensor":
        return self._tensor.__le__(other)

    def sum(
        self, axis: Optional[int] = -1, keepdims: Optional[bool] = False
    ) -> "Tensor":
        return self._tensor.sum(axis=axis, keepdims=keepdims)

    def mean(
        self, axis: Optional[int] = -1, keepdims: Optional[bool] = False
    ) -> "Tensor":
        return self._tensor.mean(axis=axis, keepdims=keepdims)

    def min(
        self, axis: Optional[int] = -1, keepdims: Optional[bool] = False
    ) -> "Tensor":
        return self._tensor.min(axis=axis, keepdims=keepdims)

    def max(
        self, axis: Optional[int] = -1, keepdims: Optional[bool] = False
    ) -> "Tensor":
        return self._tensor.max(axis=axis, keepdims=keepdims)

    def median(
        self, axis: Optional[int] = -1, keepdims: Optional[bool] = False
    ) -> "Tensor":
        return self._tensor.median(axis=axis, keepdims=keepdims)

    def backward(self, custom_grad: Optional["Tensor"] = None) -> None:
        return self._tensor.backward(custom_grad)

    def transpose(self, dim0: int, dim1: int) -> "Tensor":
        return self._tensor.transpose(dim0, dim1)

    def zero_(self) -> "Tensor": return self._tensor.zero_()
    
    def view(self, *shape) -> "Tensor": return self._tensor.view(shape)
    
    def reshape(self, *shape) -> "Tensor":
        new_data = Th.reshape(self.data, shape)
        return Tensor(new_data, self.requires_grad, self.dtype)

    def __hash__(self):
        return id(self)
    
    def tolist(self):
        return self._tensor.data
    
    def tonumpy(self):
        return array(self._tensor.data)
    
    __module__ = "anygrad"



from setuptools import setup, find_packages
from pybind11.setup_helpers import Pybind11Extension, build_ext
import sys

compile_args = ["-O2", "-std=c++20"] if sys.platform != "win32" else ["/O2", "/std:c++20"]
version = "0.0.4"

ext_modules = [
    Pybind11Extension(
        "anygrad.tensor.base.tensor_c",
        [
            "backend/anygrad/bindings/bind_tensor.cpp",
            "backend/anygrad/Th/ThAllocate.cpp",
            "backend/anygrad/Th/ThBaseops.cpp",
            "backend/anygrad/Th/Thhelpers.cpp",
            "backend/anygrad/Th/Thgemm.cpp",
            "backend/anygrad/Th/Tharrange.cpp"
        ],
        language="c++",
        extra_compile_args=compile_args
    ),
    Pybind11Extension(
        "anygrad.utils.utils_c",
        [
            "backend/anygrad/bindings/utils_bind.cpp",
            "backend/anygrad/utils/random_num.cpp",
            "backend/anygrad/Th/Thhelpers.cpp",
            "backend/anygrad/utils/init_ops.cpp",
            "backend/anygrad/utils/log_arithmetic.cpp",
        ],
        language="c++",
        extra_compile_args=compile_args
    )
]

setup(
    name="anygrad",
    version=version,
    description="A Tensor module that allows a deep learning framework to switch seamlessly between different engines.",
    long_description=open('README.md', encoding='utf-8').read(),
    long_description_content_type='text/markdown',
    author="Ruhaan",
    author_email="ruhaan123dalal@gmail.com",
    license="MIT",
    ext_modules=ext_modules,
    cmdclass={"build_ext": build_ext},
    zip_safe=False,
    packages=find_packages(),
    package_dir={"": "."},
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
    ],
    include_package_data=True,
    install_requires=[
        "pybind11", "numpy"
    ]
)



from anygrad.tensor.base.basetensor import BaseTensor
from anygrad.tensor.base.booltensor import BoolTensor
from anygrad.tensor.base.floattensor import FloatTensor
from anygrad.tensor.base.inttensor import IntTensor

__all__ = ["BaseTensor", "BoolTensor", "FloatTensor", "IntTensor"]


from contextlib import contextmanager


class GradMode:
    _enable = True

    @classmethod
    def is_enabled(cls):
        return cls._enable

    @classmethod
    def set_enable(cls, mode: bool):
        cls._enable = mode


@contextmanager
def no_grad():
    GradMode.set_enable(False)
    try:
        yield
    finally:
        GradMode.set_enable(True)


no_grad.__module__ = "anygrad.autograd"



class Version(str):
    def __init__(self, version: str):
        self.version = version

    def __repr__(self):
        return f"Version(anygrad={self.version})"

    __module__ = "anygrad"


__version__ = Version("0.0.4")



from anygrad.utils import utils_c as C
from anygrad.tensor.base import ThHelper as Th
from anygrad.tensor.tensor import Tensor
import random


class Generator(C.GeneratorBase):
    def __init__(self, seed):
        super().__init__(seed)

    def manual_seed(self, seed):
        return super().manual_seed(seed)
    
    __module__ = "anygrad"


def _use_utils_ops(operation_name, dtype, **kwargs):

    if isinstance(kwargs["shape"][0], tuple) and len(kwargs["shape"]) == 1:
        kwargs["shape"] = kwargs["shape"][0]

    dtype_mapping = {
        Th.float32: "float32",
        Th.float64: "float64",
        Th.int32: "int32",
        Th.int64: "int64",
    }
    try:
        operation_func = getattr(
            C, f"{operation_name.capitalize()}{dtype_mapping[dtype].capitalize()}"
        )
    except Exception:
        pass

    if "low" in kwargs:
        data, shape = operation_func(
            kwargs["shape"], kwargs["low"], kwargs["high"], kwargs["generator"]
        )
    else:
        data, shape = operation_func(kwargs["shape"], kwargs["generator"])
    ans = Th.reshape(data, shape)
    del data, shape
    ans = Tensor(ans, kwargs["requires_grad"], dtype=dtype_mapping[dtype])
    return ans


def rand(*shape, generator=None, requires_grad=False, dtype=Th.float32):
    if dtype not in {Th.float32, Th.float64}:
        raise TypeError(
            "In valid dtype is provide for the rand operation. use float32 and float64"
        )
    if generator is None:
        generator = C.GeneratorBase(random.randint(0, 100))
    return _use_utils_ops(
        shape=shape,
        generator=generator,
        requires_grad=requires_grad,
        operation_name="rand",
        dtype=dtype,
    )


def randint(*shape, low=0, high, generator=None, requires_grad=False, dtype=Th.int32):
    if dtype not in {Th.int32, Th.int64}:
        raise TypeError(
            "Invalid dtype is provide for the randint operation. use int32 and int64"
        )
    if generator is None:
        generator = C.GeneratorBase(random.randint(0, 100))
    return _use_utils_ops(
        shape=shape,
        low=low,
        high=high,
        generator=generator,
        requires_grad=requires_grad,
        operation_name="randint",
        dtype=dtype,
    )


rand.__module__ = "anygrad"
randint.__module__ = "anygrad"



import anygrad


class GradientCal:
    @staticmethod
    def initialize_grad(tensor):
        """Ensure tensor.grad is initialized."""
        if tensor.grad is None:
            tensor.grad = anygrad.zeros_like(tensor, requires_grad=False, dtype=tensor.dtype)

    @staticmethod
    def accumulate_grad_(tensor, factor):
        """Accumulate gradients with a scaling factor."""
        if tensor.requires_grad:
            with anygrad.no_grad():
                GradientCal.initialize_grad(tensor)
                GradientCal.accumulate_grad(tensor, factor)
 
    @staticmethod
    def accumulate_matmul(tensor, factor):
        if tensor.requires_grad:
            with anygrad.no_grad():
                GradientCal.initialize_grad(tensor)
                GradientCal.accmulate_grad_matmul(tensor, factor)

    @staticmethod
    def accmulate_grad_matmul(tensor1, tensor2):
        """Accumulate gradients for tensor1 using tensor2 for matmul."""
        tensor2.requires_grad = False
        if tensor1.grad.dtype != tensor2.dtype:
            tensor2 = anygrad.cast(tensor2, tensor1.grad.dtype)
        current = tensor2
        while current.ndim > tensor1.grad.ndim:
            current = current.sum(0)
        tensor1.grad += current

    @staticmethod
    def accumulate_grad(tensor1, tensor2):
        """Accumulate gradients for tensor1 using tensor2."""
        tensor2.requires_grad = False
        if tensor1.grad.dtype != tensor2.dtype:
            tensor2 = anygrad.cast(tensor2, tensor1.grad.dtype)
        if tensor1.grad.shape != tensor2.shape:
            tensor2 = tensor2.sum(axis=0, keepdims=True)
            tensor2.requires_grad = False
        tensor1.grad += tensor2

    @staticmethod
    def Add_grad(tensor1, tensor2, ans_tensor):
        def _backward():
            GradientCal.accumulate_grad_(tensor1, ans_tensor.grad)
            if not isinstance(tensor2, (int, float)):
                GradientCal.accumulate_grad_(tensor2, ans_tensor.grad)

        return _backward

    @staticmethod
    def Sub_grad(tensor1, tensor2, ans_tensor):
        def _backward():
            GradientCal.accumulate_grad_(tensor1, ans_tensor.grad)
            if not isinstance(tensor2, (int, float)):
                GradientCal.accumulate_grad_(tensor2, -ans_tensor.grad)

        return _backward

    @staticmethod
    def Mul_grad(tensor1, tensor2, ans_tensor):
        def _backward():
            if tensor1.requires_grad:
                GradientCal.accumulate_grad_(
                    tensor1, tensor2 * ans_tensor.grad
                )
            if not isinstance(tensor2, (int, float)) and tensor2.requires_grad:
                GradientCal.accumulate_grad_(
                    tensor2, tensor1 * ans_tensor.grad
                )

        return _backward

    @staticmethod
    def Div_grad(tensor1, tensor2, ans_tensor):
        def _backward():
            if tensor1.requires_grad:
                GradientCal.accumulate_grad_(
                    tensor1, (1 / tensor2) * ans_tensor.grad
                )
            if not isinstance(tensor2, (int, float)) and tensor2.requires_grad:
                GradientCal.accumulate_grad_(
                    tensor2, ((-tensor1) / tensor2**2) * ans_tensor.grad
                )

        return _backward

    @staticmethod
    def Pow_grad(tensor1, tensor2, ans_tensor):
        def _backward():
            if tensor1.requires_grad:
                GradientCal.accumulate_grad_(
                    tensor1, tensor2 * ans_tensor.grad * (ans_tensor / tensor1)
                )
            if not isinstance(tensor2, (int, float)) and tensor2.requires_grad:
                GradientCal.accumulate_grad_(
                    tensor2, ans_tensor.grad * ans_tensor * anygrad.log(tensor1)
                )

        return _backward

    @staticmethod
    def Matmul_grad(tensor1, tensor2, ans_tensor):
        def _backward():
            if tensor1.requires_grad:
                GradientCal.accumulate_matmul(
                    tensor1, ans_tensor.grad @ tensor2.transpose(-2, -1)
                )
            if tensor2.requires_grad:
                GradientCal.accumulate_matmul(
                    tensor2, tensor1.transpose(-2, -1) @ ans_tensor.grad
                )

        return _backward

    @staticmethod
    def Trans_grad(tensor1, ans_tensor):
        def _backward():
            if tensor1.requires_grad:
                GradientCal.accumulate_grad_(
                    tensor1, ans_tensor.grad.transpose(-2, -1)
                )

        return _backward

    @staticmethod
    def Sum_grad(tensor, ans_tensor):
        def _backward():
            if tensor.requires_grad:
                GradientCal.accumulate_grad_(tensor, ans_tensor.grad)

        return _backward

    @staticmethod
    def Mean_grad(tensor, ans_tensor):
        def _backward():
            if tensor.requires_grad:
                GradientCal.accumulate_grad_(
                    tensor, ans_tensor.grad / tensor.size
                )

        return _backward

    @staticmethod
    def Median_grad(tensor, ans_tensor):
        def _backward():
            if tensor.requires_grad:
                median_val = (
                    ans_tensor.data[0]
                    if isinstance(ans_tensor.data, list) and len(ans_tensor.data) == 1
                    else ans_tensor.data
                )
                mask = anygrad.Tensor(
                    [1 if x == median_val else 0 for x in tensor.data]
                )
                count = sum(mask) or 1
                grad_factor_tensor = ans_tensor.grad * (mask / count)
                GradientCal.accumulate_grad_(tensor, grad_factor_tensor)

        return _backward

    @staticmethod
    def Max_grad(tensor, ans_tensor):
        def _backward():
            if tensor.requires_grad:
                GradientCal.accumulate_grad_(tensor, ans_tensor.grad)

        return _backward

    @staticmethod
    def Min_grad(tensor, ans_tensor):
        def _backward():
            if tensor.requires_grad:
                GradientCal.accumulate_grad_(tensor, ans_tensor.grad)

        return _backward

    @staticmethod
    def View_grad(tensor, ans_tensor):
        def _backward():
            if tensor.requires_grad:
                GradientCal.accumulate_grad_(tensor, ans_tensor.grad)
                
        return _backward
                
    @staticmethod
    def Reshape_grad(tensor, ans_tensor):
        def _backward():
            if tensor.requires_grad:
                GradientCal.accumulate_grad_(tensor, ans_tensor.grad)
                
        return _backward
                
            


from typing import Optional, List, Union
import pprint

from anygrad.tensor.base import tensor_c as C
from anygrad.tensor.base import ThHelper as Th
from anygrad.tensor.base.basetensor import BaseTensor
from anygrad.tensor.base.booltensor import BoolTensor


class FloatTensor(BaseTensor):
    """
    Class to repesent a FloatTensor.

    Attributes
    ----------
    data : List | Tuple
        Any Iterable

    requires_grad: Optional[bool] = False
        if `True` the gradient caculation is happend

    dtype: Optional[anygrad.float32 | anygrad.float64] = anygrad.float32

    Methods
    ----------
    data:
        return the item of the tensor in list form.
    shape:
        return the shape of the tensor in tuple form.
    ndim:
        return the dim of the tensor in int form.
    requires_grad:
        return the bool value if the requires_grad.
    grad:
        a tensor value that allow you to see the gredient of the tensor.

    add(other):
        other: Tensor | int | float
        add the Tensor or number.

    sub(other):
        other: Tensor | int | float
        sub the Tensor or number.

    mul(other):
        other: Tensor | int | float
        mul the Tensor or number.

    div(other):
        other: Tensor | int | float
        div the Tensor or number.

    pow(other):
        other: Tensor | int | float
        pow the Tensor or number.

    matmul(other):
        other: Tensor
        matrix multiplication of the Two valid shape tensor.

    sum(self, axis: Optional[int] = -1, keepdims: Optional[bool] = False) -> Tensor:
        sum the tensor with axis and keepdims.

    backward(self, custom_grad:Optional[Tensor] = None) -> None:
        Do the backward pass if the requires-grad is true for given tensor.

    """

    def __init__(
        self,
        data: List[int],
        requires_grad=True,
        dtype: Optional[Th.float32 | Th.float64] = Th.float32,
    ):
        super().__init__(requires_grad)

        if isinstance(data, (int, float)):
            data = [data]

        list_data = Th.flat_list(data)

        if isinstance(dtype, str):
            dtype = getattr(Th, dtype)

        self.data = Th.convert_tensor(data, float)
        shape = Th.cal_shape(data)

        if dtype == Th.float32:
            self.base = C.float32(list_data, shape)
        elif dtype == Th.float64:
            self.base = C.float64(list_data, shape)

    def __repr__(self):
        data = Th.round_list(self.data)
        formate_data = pprint.pformat(data, width=150, depth=50)
        base_str = f"Tensor({formate_data}"

        if self.name_backward:
            return base_str + f" name_backward = {self.name_backward})"

        if self.requires_grad:
            return base_str + f" requires_grad = {self.requires_grad})"

        if self.base.dtype != "float32":
            return base_str + f" dtype= {self.dtype})"
        return base_str + ")"

    def __getitem__(self, index: Union[int, slice]) -> "FloatTensor":
        new_data = self.data[index]
        return FloatTensor(new_data, requires_grad=self.requires_grad, dtype=self.dtype)

    def __add__(self, other) -> "FloatTensor":
        return BaseTensor._apply_operation(
            tensor1=self,
            tensor2=other,
            ResultClass=FloatTensor,
            OtherClass=FloatTensor,
            has_scaler=True,
            operation=lambda x, y: x + y,
            operation_name="Add",
            allow_other_class=False
        )

    def __radd__(self, other) -> "FloatTensor":
        return self._add__(other)

    def __sub__(self, other) -> "FloatTensor":
        return BaseTensor._apply_operation(
            tensor1=self,
            tensor2=other,
            ResultClass=FloatTensor,
            OtherClass=FloatTensor,
            has_scaler=True,
            operation=lambda x, y: x - y,
            operation_name="Sub",
            allow_other_class=False
        )

    def __rsub__(self, other) -> "FloatTensor":
        return BaseTensor._apply_operation(
            tensor1=self,
            tensor2=other,
            ResultClass=FloatTensor,
            OtherClass=FloatTensor,
            has_scaler=True,
            operation=lambda x, y: y - x,
            operation_name="Sub",
            allow_other_class=False
        )

    def __mul__(self, other) -> "FloatTensor":
        return BaseTensor._apply_operation(
            tensor1=self,
            tensor2=other,
            ResultClass=FloatTensor,
            OtherClass=FloatTensor,
            has_scaler=True,
            operation=lambda x, y: x * y,
            operation_name="Mul",
            allow_other_class=False
        )

    def __rmul__(self, other) -> "FloatTensor":
        return self.__mul__(other)

    def __truediv__(self, other) -> "FloatTensor":
        return BaseTensor._apply_operation(
            tensor1=self,
            tensor2=other,
            ResultClass=FloatTensor,
            OtherClass=FloatTensor,
            has_scaler=True,
            operation=lambda x, y: x / y,
            operation_name="Div",
            allow_other_class=True
        )

    def __rtruediv__(self, other) -> "FloatTensor":
        return BaseTensor._apply_operation(
            tensor1=self,
            tensor2=other,
            ResultClass=FloatTensor,
            OtherClass=FloatTensor,
            has_scaler=True,
            operation=lambda x, y: y / x,
            operation_name="Div",
            allow_other_class=True
        )

    def __pow__(self, other) -> "FloatTensor":
        return BaseTensor._apply_operation(
            tensor1=self,
            tensor2=other,
            ResultClass=FloatTensor,
            OtherClass=FloatTensor,
            has_scaler=True,
            operation=lambda x, y: x ** y,
            operation_name="Pow",
            allow_other_class=False
        )

    def __matmul__(self, other) -> "FloatTensor":
        return BaseTensor._apply_operation(
            tensor1=self,
            tensor2=other,
            ResultClass=FloatTensor,
            OtherClass=FloatTensor,
            has_scaler=False,
            operation=None,
            operation_name="Matmul",
            allow_other_class=False,
            broadcast_check=C.is_matmul_broadcast
        )
        
    def __eq__(self, other):
        return BaseTensor._apply_compare(
            self,
            other,
            BoolTensor,
            operation = lambda x, y: x == y,
            operation_name = "Eq",
            has_scaler=True
        )
    
    def __gt__(self, other):
        return BaseTensor._apply_compare(
            self,
            other,
            BoolTensor,
            operation = lambda x, y: x > y,
            operation_name = "Gt",
            has_scaler=True
        )
        
    def __lt__(self, other):
        return BaseTensor._apply_compare(
            self,
            other,
            BoolTensor,
            operation = lambda x, y: x < y,
            operation_name = "Lt",
            has_scaler=True
        )
    
    def __ge__(self, other):
        return BaseTensor._apply_compare(
            self,
            other,
            BoolTensor,
            operation = lambda x, y: x >= y,
            operation_name = "Ge",
            has_scaler=True
        )
    
    def __le__(self, other):
        return BaseTensor._apply_compare(
            self,
            other,
            BoolTensor,
            operation = lambda x, y: x <= y,
            operation_name = "Le",
            has_scaler=True
        )
        

    def sum(self, axis: Optional[int] = -1, keepdims: Optional[bool] = False):
        return BaseTensor._reduce_ops(self, FloatTensor, FloatTensor, axis, keepdims, "Sum", allow_other_class=False)

    def mean(self, axis: Optional[int] = -1, keepdims: Optional[bool] = False):
        return BaseTensor._reduce_ops(self, FloatTensor, FloatTensor, axis, keepdims, "Mean", allow_other_class=False)

    def min(self, axis: Optional[int] = -1, keepdims: Optional[bool] = False):
        return BaseTensor._reduce_ops(self, FloatTensor, FloatTensor, axis, keepdims, "Min", allow_other_class=False)

    def max(self, axis: Optional[int] = -1, keepdims: Optional[bool] = False):
        return BaseTensor._reduce_ops(self, FloatTensor, FloatTensor, axis, keepdims, "Max", allow_other_class=False)

    def median(self, axis: Optional[int] = -1, keepdims: Optional[bool] = False):
        return BaseTensor._reduce_ops(self, FloatTensor, FloatTensor, axis, keepdims, "Median", allow_other_class=False)

    def transpose(self, dim0: int, dim1: int) -> "FloatTensor":
        return BaseTensor._trans_ops(self, FloatTensor, FloatTensor, dim0, dim1, allow_other_class=False)

    def zero_(self) -> "FloatTensor":
        self.data = [[0.0] * dim for dim in self.shape]
        self.base = C.float32(Th.flat_list(self.data), self.shape) if self.dtype == "float32" else C.float64(Th.flat_list(self.data), self.shape)
        return self

    def view(self, shape) -> "FloatTensor":
        return BaseTensor._apply_view(self, FloatTensor, FloatTensor, shape, allow_other_class=False)
    
    def __hash__(self):
        return id(self)

    __module__ = "anygrad"
